package org.topbraid.spin.arq;

import java.util.List;

import org.topbraid.spin.model.Function;
import org.topbraid.spin.model.Query;
import org.topbraid.spin.system.ARQFactory;
import org.topbraid.spin.vocabulary.SPIN;

import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QuerySolution;
import com.hp.hpl.jena.query.QuerySolutionMap;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.sparql.core.Var;
import com.hp.hpl.jena.sparql.engine.binding.Binding;
import com.hp.hpl.jena.sparql.expr.Expr;
import com.hp.hpl.jena.sparql.expr.ExprList;
import com.hp.hpl.jena.sparql.expr.NodeValue;
import com.hp.hpl.jena.sparql.function.FunctionEnv;
import com.hp.hpl.jena.sparql.function.FunctionFactory;


/**
 * An ARQ function that delegates its functionality into a user-defined
 * SPIN function. 
 * 
 * @author Holger Knublauch
 */
public class SPINARQFunction implements com.hp.hpl.jena.sparql.function.Function, FunctionFactory {
	
	private com.hp.hpl.jena.query.Query arqQuery;
	
	private String queryString;
	

	/**
	 * Constructs a new SPINARQFunction based on a given SPIN Function.
	 * The spinFunction model be associated with the Model containing
	 * the triples of its definition.
	 * @param spinFunction  the SPIN function
	 */
	public SPINARQFunction(Function spinFunction) {
		try {
			Query spinQuery = (Query) spinFunction.getBody();
			queryString = spinQuery.toString();
			arqQuery = ARQFactory.get().createQuery(spinFunction.getModel(), queryString);
		}
		catch(Throwable t) {
			throw new IllegalArgumentException("Function definition does not contain a valid body", t);
		}
	}
	

	public void build(String uri, ExprList args) {
	}

	
	public com.hp.hpl.jena.sparql.function.Function create(String uri) {
		return this;
	}

	
	public NodeValue exec(Binding binding, ExprList args, String uri, FunctionEnv env) {
		Model model = ModelFactory.createModelForGraph(env.getActiveGraph());
		QueryExecution qexec = ARQFactory.get().createQueryExecution(arqQuery, model);
		Node t = binding.get(Var.alloc(SPIN.THIS_VAR_NAME));
		QuerySolutionMap bindings = new QuerySolutionMap();
		if(t != null) {
			bindings.add(SPIN.THIS_VAR_NAME, model.asRDFNode(t));
		}
		for(int i = 0; i < args.size(); i++) {
			Expr expr = args.get(i);
			if(!expr.isVariable() || binding.contains(expr.asVar())) {
	        	NodeValue x = expr.eval(binding, env);
	        	if(x != null) {
	        		bindings.add("arg" + (i + 1), model.asRDFNode(x.asNode()));
	        	}
			}
		}
		qexec.setInitialBinding(bindings);
		if(arqQuery.isAskType()) {
			boolean result = qexec.execAsk();
			qexec.close();
			return NodeValue.makeBoolean(result);
		}
		else if(arqQuery.isSelectType()) {
			ResultSet rs = qexec.execSelect();
			if(rs.hasNext()) {
				QuerySolution s = rs.nextSolution();
				List<String> resultVars = rs.getResultVars();
				String varName = resultVars.get(0);
				RDFNode resultNode = s.get(varName);
				qexec.close();
				if(resultNode != null) {
					return NodeValue.makeNode(resultNode.asNode());
				}
			}
			else {
				qexec.close();
			}
			return null;
		}
		else {
			return null;
		}
	}
}
