package org.topbraid.spin.model;

import java.util.Iterator;

import org.topbraid.spin.model.impl.TriplePatternImpl;
import org.topbraid.spin.model.update.Delete;
import org.topbraid.spin.model.update.Insert;
import org.topbraid.spin.model.update.Modify;
import org.topbraid.spin.model.update.Update;
import org.topbraid.spin.system.SPINModuleRegistry;
import org.topbraid.spin.util.JenaUtil;
import org.topbraid.spin.vocabulary.SP;
import org.topbraid.spin.vocabulary.SPIN;
import org.topbraid.spin.vocabulary.SPL;

import com.hp.hpl.jena.enhanced.EnhGraph;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFList;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.vocabulary.RDF;


/**
 * The singleton that is used to convert plain Jena objects into
 * SPIN API resources, and to do corresponding tests.
 * 
 * @author Holger Knublauch
 */
public class SPINFactory {
	
	/**
	 * Attempts to cast a given Resource into the most specific
	 * subclass of Command, esp Update or Query.
	 * @param resource  the Resource to cast
	 * @return resource cast into the best possible type or null
	 */
	public static Command asCommand(Resource resource) {
		Query query = asQuery(resource);
		if(query != null) {
			return query;
		}
		else {
			return asUpdate(resource);
		}
	}

	
	/**
	 * Checks whether a given Resource represents a SPARQL element, and returns
	 * an instance of a subclass of Element if so.
	 * @param resource  the Resource to check
	 * @return Resource as an Element or null if resource is not an element
	 */
	public static Element asElement(Resource resource) {
		final TriplePattern triplePattern = asTriplePattern(resource);
		if(triplePattern != null) {
			return triplePattern;
		}
		else if(resource.canAs(TriplePath.class)) {
			return (TriplePath) resource.as(TriplePath.class);
		}
		else if(resource.canAs(Filter.class)) {
			return (Filter) resource.as(Filter.class);
		}
		else if(resource.canAs(Let.class)) {
			return (Let) resource.as(Let.class);
		}
		else if(resource.canAs(Optional.class)) {
			return (Optional) resource.as(Optional.class);
		}
		else if(resource.canAs(NamedGraph.class)) {
			return (NamedGraph) resource.as(NamedGraph.class); 
		}
		else if(resource.canAs(NotExists.class)) {
			return (NotExists) resource.as(NotExists.class);
		}
		else if(resource.canAs(Service.class)) {
			return (Service) resource.as(Service.class);
		}
		else if(resource.canAs(SubQuery.class)) {
			return (SubQuery) resource.as(SubQuery.class); 
		}
		else if(resource.canAs(Union.class)) {
			return (Union) resource.as(Union.class); 
		}
		else if(isElementList(resource)) {
			return (ElementList) resource.as(ElementList.class);
		}
		else {
			return null;
		}
	}

	
	/**
	 * Returns the most specific Java instance for a given RDFNode.
	 * If the node is a function call, it will be returned as instance of
	 * Function.  If it's a Variable, the Variable will be returned.
	 * Otherwise the node itself will be returned.
	 * @param node  the node to cast
	 * @return node or node as a Function or Variable
	 */
	public static RDFNode asExpression(RDFNode node) {
		if(node instanceof Resource) {
			Variable var = SPINFactory.asVariable(node);
			if(var != null) {
				return var;
			}
			FunctionCall function = SPINFactory.asFunctionCall((Resource)node);
			if(function != null) {
				return function;
			}
		}
		return node;
	}
	
	
	/**
	 * Converts a given Resource into a Function instance.
	 * No other tests are done.
	 * @param resource  the Resource to convert
	 * @return the Function
	 */
	public static Function asFunction(Resource resource) {
		return (Function) resource.as(Function.class);
	}

	
	/**
	 * Checks if a given Resource might represent a Function call, and if
	 * yes returns the resource as Function.  The condition here is fairly
	 * general: a function call must be a blank node with an rdf:type triple
	 * where the type triple's object is a URI resource.  It is generally
	 * assumed that this function is called after other options have been
	 * exhausted.  For example, in order to test whether a resource is a
	 * variable or a function call, the variable test must be done first
	 * as it is more specific than these test conditions 
	 * @param resource  the Resource to test
	 * @return resource as a Function or null if resource cannot be cast
	 */
	public static FunctionCall asFunctionCall(Resource resource) {
		if(resource.isAnon()) {
			Statement s = resource.getProperty(RDF.type);
			if(s != null && s.getObject().isURIResource()) {
				return (FunctionCall) resource.as(FunctionCall.class);
			}
		}
	    return null;
	}

	
	/**
	 * Checks if a given Resource is a SPIN query, and returns an
	 * instance of a subclass of Query if so. 
	 * @param resource  the Resource to test
	 * @return resource as a Query or null
	 */
	public static Query asQuery(Resource resource) {
		if(resource.canAs(Select.class)) {
			return (Select) resource.as(Select.class);
		}
		else if(resource.canAs(Construct.class)) {
			return (Construct) resource.as(Construct.class);
		}
		else if(resource.canAs(Ask.class)) {
			return (Ask) resource.as(Ask.class);
		}
		else if(resource.canAs(Describe.class)) {
			return (Describe) resource.as(Describe.class);
		}
		else {
			return null;
		}
	}
	
	
	/**
	 * Converts a given Resource into a Template instance.
	 * No other tests are done.
	 * @param resource  the Resource to convert
	 * @return the Template
	 */
	public static Template asTemplate(Resource resource) {
		return (Template) resource.as(Template.class);
	}
	
	
	/**
	 * Checks whether a given RDFNode can be cast into TemplateCall, and returns
	 * it as a TemplateCall instance if so.
	 * @param node  the node to convert
	 * @return an instance of TemplateCall or null
	 */
	public static TemplateCall asTemplateCall(RDFNode node) {
		if(node instanceof Resource) {
			Statement s = ((Resource)node).getProperty(RDF.type);
			if(s != null && s.getObject().isURIResource()) {
				String uri = s.getResource().getURI();
				Template template = SPINModuleRegistry.get().getTemplate(uri, s.getModel());
				if(template != null) {
					return (TemplateCall) node.as(TemplateCall.class);
				}
			}
		}
		return null;
	}

	
	/**
	 * Checks whether a given RDFNode can be converted into a TriplePattern, and if yes,
	 * returns an instance of TriplePattern.
	 * @param node  the node to test
	 * @return node as TriplePattern or null
	 */
	public static TriplePattern asTriplePattern(RDFNode node) {
		if(node instanceof Resource && ((Resource)node).hasProperty(SP.predicate)) {
			return new TriplePatternImpl(node.asNode(), (EnhGraph)((Resource)node).getModel());
		}
		else {
			return null;
		}
	}
	
	
	/**
	 * Checks if a given Resource is a subclass of sp:Update and
	 * casts it into the most specific Java class possible.
	 * @param resource  the Resource to cast
	 * @return the Update or null if resource cannot be cast
	 */
	public static Update asUpdate(Resource resource) {
		if(resource.canAs(Modify.class)) {
			return (Modify) resource.as(Modify.class);
		}
		else if(resource.canAs(Delete.class)) {
			return (Delete) resource.as(Delete.class);
		}
		else if(resource.canAs(Insert.class)) {
			return (Insert) resource.as(Insert.class);
		}
		else {
			return null;
		}
	}
	
	
	/**
	 * Checks whether a given RDFNode can be cast into a Variable and - if yes -
	 * converts it into an instance of Variable.  The Resource must have a value
	 * for spin:varName.
	 * @param node  the node to check
	 * @return resource as a Variable or null
	 */
	public static Variable asVariable(RDFNode node) {
		if(node instanceof Resource && ((Resource)node).hasProperty(SP.varName)) {
			return (Variable)node.as(Variable.class);
		}
		else {
			return null;
		}
	}
	

	/**
	 * Creates an spl:Argument with a given property and value type.
	 * The new Argument resource will be a blank node in a given Model.
	 * @param model  the Model
	 * @param argProperty  the property or null
	 * @param argType  the value type or null
	 * @param optional  true if the Argument shall be optional
	 * @return the new Argument
	 */
	public static Argument createArgument(Model model, Property argProperty, Resource argType, boolean optional) {
		Argument a = (Argument) model.createResource(SPL.Argument).as(Argument.class);
		if(argProperty != null) {
			a.addProperty(SPL.predicate, argProperty);
		}
		if(argType != null) {
			a.addProperty(SPL.valueType, argType);
		}
		if(optional) {
			a.addProperty(SPL.optional, model.createTypedLiteral(true));
		}
		return a;
	}
	

	/**
	 * Creates a new spl:Attribute as a blank node in a given Model.
	 * @param model  the Model to create the attribute in
	 * @param argProperty  the predicate or null
	 * @param argType  the value type or null
	 * @param minCount  the minimum cardinality or null
	 * @param maxCount  the maximum cardinality or null
	 * @return a new Attribute
	 */
	public static Attribute createAttribute(Model model, Property argProperty, Resource argType, Integer minCount, Integer maxCount) {
		Attribute a = (Attribute) model.createResource(SPL.Attribute).as(Attribute.class);
		if(argProperty != null) {
			a.addProperty(SPL.predicate, argProperty);
		}
		if(argType != null) {
			a.addProperty(SPL.valueType, argType);
		}
		if(minCount != null) {
			a.addProperty(SPL.minCount, model.createTypedLiteral(minCount.intValue()));
		}
		if(maxCount != null) {
			a.addProperty(SPL.maxCount, model.createTypedLiteral(maxCount.intValue()));
		}
		return a;
	}
	
	
	/**
	 * Creates an Ask query for a given WHERE clause.
	 * @param model  the Model to create the Ask (blank node) in
	 * @param where  the elements of the WHERE clause
	 * @return the new Ask query
	 */
	public static Ask createAsk(Model model, ElementList where) {
		Ask ask = (Ask) model.createResource(SP.Ask).as(Ask.class);
		ask.addProperty(SP.where, where);
		return ask;
	}
	

	/**
	 * Creates a new ElementList in a given Model.
	 * @param model  the Model to create the ElementList in
	 * @param elements  the elements (may be empty)
	 * @return a new ElementList (may be rdf:nil)
	 */
	public static ElementList createElementList(Model model, Element[] elements) {
		if(elements.length > 0) {
			return (ElementList) model.createList(elements).as(ElementList.class);
		}
		else {
			return (ElementList) RDF.nil.inModel(model).as(ElementList.class);
		}
	}
	
	
	/**
	 * Creates a new ElementList in a given Model.
	 * @param model  the Model to create the ElementList in
	 * @param elements  the elements (may be empty)
	 * @return a new ElementList (may be rdf:nil)
	 */
	public static ElementList createElementList(Model model, Iterator<Element> elements) {
		if(elements.hasNext()) {
			return (ElementList) model.createList(elements).as(ElementList.class);
		}
		else {
			return (ElementList) RDF.nil.inModel(model).as(ElementList.class);
		}
	}


	/**
	 * Creates a Filter from a given expression.
	 * @param model  the Model to create the (blank node) Filter in
	 * @param expression  the expression node (not null)
	 * @return a new Filter
	 */
	public static Filter createFilter(Model model, RDFNode expression) {
		Filter filter = (Filter) model.createResource(SP.Filter).as(Filter.class);
		filter.addProperty(SP.expression, expression);
		return filter;
	}
	

	/**
	 * Creates a new Function call, which is basically an instance of the
	 * function's class.
	 * @param model  the Model to create the function call in
	 * @param function  the function class (must be a URI resource)
	 * @return a new instance of function
	 */
	public static FunctionCall createFunctionCall(Model model, Resource function) {
		return (FunctionCall) model.createResource(function).as(FunctionCall.class);
	}
	

	/**
	 * Creates a Let in a given Model as a blank node.
	 * @param model  the Model to create the Let in
	 * @param variable  the Variable to assign
	 * @param expression  the expression on the right hand side
	 * @return a new Let instance
	 */
	public static Let createLet(Model model, Variable variable, RDFNode expression) {
		Let let = (Let) model.createResource(SP.Let).as(Let.class);
		if(variable != null) {
			let.addProperty(SP.variable, variable);
		}
		if(expression != null) {
			let.addProperty(SP.expression, expression);
		}
		return let;
	}


	/**
	 * Creates a new NamedGraph element as a blank node in a given Model.
	 * @param model  the Model to generate the NamedGraph in
	 * @param graphNameNode  the URI resource of the graph name
	 * @param elements  the elements in the NamedGraph
	 * @return a new NamedGraph
	 */
	public static NamedGraph createNamedGraph(Model model, Resource graphNameNode, RDFList elements) {
		NamedGraph result = (NamedGraph) model.createResource(SP.NamedGraph).as(NamedGraph.class);
		result.addProperty(SP.graphNameNode, graphNameNode);
		result.addProperty(SP.elements, elements);
		return result;
	}
	
	
	/**
	 * Creates a new NotExists as a blank node in a given Model.
	 * @param model  the Model to create the NOT EXISTS in
	 * @param elements  the elements of the NOT EXIST
	 * @return a new NotExists
	 */
	public static NotExists createNotExists(Model model, ElementList elements) {
		NotExists notExists = (NotExists) model.createResource(SP.NotExists).as(NotExists.class);
		notExists.addProperty(SP.elements, elements);
		return notExists;
	}
	

	/**
	 * Creates a new Optional as a blank node in a given Model. 
	 * @param model  the Model to create the OPTIONAL in
	 * @param elements  the elements of the OPTIONAL
	 * @return a new Optional
	 */
	public static Optional createOptional(Model model, ElementList elements) {
		Optional optional = (Optional) model.createResource(SP.Optional).as(Optional.class);
		optional.addProperty(SP.elements, elements);
		return optional;
	}
	
	
	public static Service createService(Model model, Resource serviceURI, ElementList elements) {
		Service service = (Service) model.createResource(SP.Service).as(Service.class);
		service.addProperty(SP.serviceURI, serviceURI);
		service.addProperty(SP.elements, elements);
		return service;
	}
	

	/**
	 * Creates a new SubQuery as a blank node in a given Model.
	 * @param model  the Model to create the SubQuery in
	 * @param subQuery  the nested query
	 * @return a new SubQuery
	 */
	public static SubQuery createSubQuery(Model model, Query subQuery) {
		SubQuery result = (SubQuery) model.createResource(SP.SubQuery).as(SubQuery.class);
		result.addProperty(SP.query, subQuery);
		return result;
	}
	
	
	/**
	 * Creates a new TemplateCall as a blank node instance of a given template.
	 * @param model  the Model to create a template call in
	 * @param template  the template class
	 * @return the new TemplateCall or null
	 */
	public static TemplateCall createTemplateCall(Model model, Resource template) {
		TemplateCall templateCall = (TemplateCall) model.createResource(template).as(TemplateCall.class);
		return templateCall;
	}


	/**
	 * Creates a new TriplePath as a blank node in a given Model.
	 * @param model  the Model to create the path in
	 * @param subject  the subject (not null)
	 * @param path  the path (not null)
	 * @param object  the object (not null)
	 * @return a new TriplePath
	 */
	public static TriplePath createTriplePath(Model model, Resource subject, Resource path, RDFNode object) {
		TriplePath triplePath = (TriplePath) model.createResource(SP.TriplePath).as(TriplePath.class);
		triplePath.addProperty(SP.subject, subject);
		triplePath.addProperty(SP.path, path);
		triplePath.addProperty(SP.object, object);
		return triplePath;
	}


	/**
	 * Creates a new TriplePattern as a blank node in a given Model.
	 * @param model  the Model to create the pattern in
	 * @param subject  the subject (not null)
	 * @param predicate  the predicate (not null)
	 * @param object  the object (not null)
	 * @return a new TriplePattern
	 */
	public static TriplePattern createTriplePattern(Model model, Resource subject, Resource predicate, RDFNode object) {
		// No rdf:type sp:TriplePattern needed - engine looks for sp:predicate
		TriplePattern triplePattern = (TriplePattern) model.createResource().as(TriplePattern.class);
		triplePattern.addProperty(SP.subject, subject);
		triplePattern.addProperty(SP.predicate, predicate);
		triplePattern.addProperty(SP.object, object);
		return triplePattern;
	}
	

	/**
	 * Creates a new UNION element as a blank node in a given Model.
	 * @param model  the Model to create the Union in
	 * @param elements  the elements
	 * @return a new Union
	 */
	public static Union createUnion(Model model, ElementList elements) {
		Union union = (Union) model.createResource(SP.Union).as(Union.class);
		union.addProperty(SP.elements, elements);
		return union;
	}
	
	
	/**
	 * Creates a new Variable as a blank node in a given Model.
	 * @param model  the Model
	 * @param varName  the name of the variable
	 * @return the Variable
	 */
	public static Variable createVariable(Model model, String varName) {
		Variable variable = (Variable) model.createResource().as(Variable.class);
		variable.addProperty(SP.varName, model.createTypedLiteral(varName));
		return variable;
	}
	
	
	/**
	 * Gets an spl:Attribute defined for a given property on a given class.
	 * The spl:Attribute must be a direct spin:constraint on the class.
	 * @param cls  the class
	 * @param property  the property
	 * @return the Attribute or null if none is found
	 */
	public static Attribute getAttribute(Resource cls, Property property) {
		StmtIterator it = JenaUtil.listAllProperties(cls, SPIN.constraint);
		while(it.hasNext()) {
			RDFNode object = it.nextStatement().getObject();
			if(object instanceof Resource && ((Resource)object).hasProperty(RDF.type, SPL.Attribute)) {
				Attribute a = (Attribute) object.as(Attribute.class);
				if(property.equals(a.getPredicate())) {
					it.close();
					return a;
				}
			}
		}
		return null;
	}
	
	
	/**
	 * Gets the most appopriate metaclass to wrap a given Command into a
	 * Template.  For example, for an Ask query, this will return spin:AskTemplate.
	 * @param command  the Command, cast into the best possible subclass
	 * @return the Template metaclass
	 */
	public static Resource getTemplateMetaClass(Command command) {
		if(command instanceof Ask) {
			return SPIN.AskTemplate;
		}
		else if(command instanceof Construct) {
			return SPIN.ConstructTemplate;
		}
		else if(command instanceof Select) {
			return SPIN.SelectTemplate;
		}
		else if(command instanceof Update) {
			return SPIN.UpdateTemplate;
		}
		else {
			throw new IllegalArgumentException("Unsupported Command type: " + command.getClass());
		}
	}
	

	/**
	 * Checks whether a given module has been declared abstract using
	 * <code>spin:abstract</code.
	 * @param module  the module to test
	 * @return true if abstract
	 */
	public static boolean isAbstract(Resource module) {
		return module.hasProperty(SPIN.abstract_, module.getModel().createTypedLiteral(true));
	}


	/**
	 * Checks if a given Resource can be cast into an ElementList.
	 * It must be either rdf:nil or an rdf:List where the first
	 * list item is an element using <code>asElement()</code>.
	 * @param resource  the resource to test
	 * @return true if resource is an element list
	 */
	public static boolean isElementList(Resource resource) {
		if(RDF.nil.equals(resource)) {
			return true;
		}
		else {
			Statement firstS = resource.getProperty(RDF.first);
			if(firstS != null && firstS.getObject().isResource()) {
				Resource first = firstS.getResource();
				return asElement(first) != null;
			}
			else {
				return false;
			}
		}
	}
	
	
	/**
	 * Checks if a given Resource is an instance of a class that has
	 * type spin:Module (or its subclasses such as spin:Function).
	 * @param resource  the Resource to check
	 * @return true  if resource is a Module
	 */
	public static boolean isModuleInstance(Resource resource) {
		Resource moduleClass = (Resource) SPIN.Module.inModel(resource.getModel());
		for(Resource type : JenaUtil.getTypes(resource)) {
			if(JenaUtil.hasIndirectType(type, moduleClass)) {
				return true;
			}
		}
		return false;
	}


	/**
	 * Checks if a given Property is spin:query or a sub-property of it.
	 * @param predicate  the Property to test
	 * @return true if predicate is a query property
	 */
	public static boolean isQueryProperty(Property predicate) {
		return SPIN.query.equals(predicate) || JenaUtil.hasSuperProperty(predicate, SPIN.query);
	}
	
	
	/**
	 * Checks whether a given RDFNode is a TemplateCall.  The condition for this
	 * is stricter than for <code>asTemplateCall</code> as the node also must have
	 * a valid template assigned to it, i.e. the type of the node must be an
	 * instance of spin:Template.
	 * @param node  the RDFNode to check
	 * @return true if node is a TemplateCall
	 */
	public static boolean isTemplateCall(RDFNode node) {
		TemplateCall templateCall = asTemplateCall(node);
		return templateCall != null && templateCall.getTemplate() != null;
	}
}
