package org.topbraid.spin.model.impl;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.topbraid.spin.model.FunctionCall;
import org.topbraid.spin.model.SPINFactory;
import org.topbraid.spin.model.print.PrintContext;
import org.topbraid.spin.system.SPINModuleRegistry;
import org.topbraid.spin.util.JenaUtil;
import org.topbraid.spin.vocabulary.SP;
import org.topbraid.spin.vocabulary.SPIN;

import com.hp.hpl.jena.enhanced.EnhGraph;
import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.RDFS;


public class FunctionCallImpl extends ModuleCallImpl implements FunctionCall {
    
	public FunctionCallImpl(Node node, EnhGraph graph) {
		super(node, graph);
	}
	
	
	public List<RDFNode> getArguments() {
		List<RDFNode> results = new LinkedList<RDFNode>();
		for(int i = 1; ; i++) {
			Property predicate = getModel().createProperty(SP.NS + "arg" + i);
			Statement s = getProperty(predicate);
			if(s != null) {
				RDFNode node = (RDFNode) s.getObject();
				results.add(SPINFactory.asExpression(node));
			}
			else {
				break;
			}
		}
		return results;
	}
	
	
	public Map<Property, RDFNode> getArgumentsMap() {
		Map<Property,RDFNode> results = new LinkedHashMap<Property, RDFNode>();
		for(int i = 1; ; i++) {
			Property predicate = getModel().createProperty(SP.NS + "arg" + i);
			Statement s = getProperty(predicate);
			if(s != null) {
				RDFNode node = (RDFNode) s.getObject();
				results.put(predicate, SPINFactory.asExpression(node));
			}
			else {
				break;
			}
		}
		return results;
	}


	private void printFunctionQName(PrintContext p, Resource function) {
		String symbol = getSymbol(function);
		if(symbol != null) {
			p.print(symbol);
		}
		else {
			String uri = function.getURI();
			p.printURIResource(getModel().getResource(uri));
		}
	}
	
	
	public Resource getFunction() {
		
		// Need to iterate over rdf:types - some may have been inferred
		// Return the most specific type, i.e. the one that does not have
		// any subclasses
		Resource type = null;
		StmtIterator it = listProperties(RDF.type);
		while(it.hasNext()) {
			Statement s = it.nextStatement();
			if(s.getObject().isURIResource()) {
				Resource candidate = s.getResource();
				if(type == null) {
					type = candidate;
				}
				else if(!getModel().contains(null, RDFS.subClassOf, candidate)) {
					type = candidate;
				}
			}
		}
		
		if(type != null) {
			if(JenaUtil.hasIndirectType(type, (Resource)SPIN.Function.inModel(type.getModel()))) {
				return type;
			}
			else {
				Resource global = SPINModuleRegistry.get().getFunction(type.getURI(), null);
				if(global != null) {
					return global;
				}
				else {
					return type;
				}
			}
		}
		else {
			return null;
		}
	}
	
	
	private String getSymbol(Resource function) {
		if(function != null) {
			Statement s = function.getProperty(SPIN.symbol);
			if(s != null && s.getObject().isLiteral()) {
				return s.getString();
			}
		}
		return null;
	}
	
	
	public void print(PrintContext p) {
		Resource function = getFunction();
		List<RDFNode> args = getArguments();
		
		String symbol = getSymbol(function);
		if(symbol != null && !Character.isLetter(symbol.charAt(0))) {
			if(p.isNested()) {
				p.print("(");
			}
			if(args.size() == 1) {
				p.print(symbol);
				printNestedExpressionString(p, args.get(0));
			}
			else { // assuming params.size() == 2
				printNestedExpressionString(p, args.get(0));
				p.print(" ");
				p.print(symbol);
				p.print(" ");
				printNestedExpressionString(p, args.get(1));
			}
			if(p.isNested()) {
				p.print(")");
			}
		}
		else {
			printFunctionQName(p, function);
			p.print("(");
			Iterator<RDFNode> it = args.iterator();
			while(it.hasNext()) {
				RDFNode param = it.next();
				printNestedExpressionString(p, param);
				if(it.hasNext()) {
					p.print(", ");
				}
			}
			p.print(")");
		}
	}
}
