package org.topbraid.spin.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.topbraid.spin.internal.ContainsVarChecker;
import org.topbraid.spin.model.Query;
import org.topbraid.spin.model.QueryOrTemplateCall;
import org.topbraid.spin.model.SPINFactory;
import org.topbraid.spin.model.Template;
import org.topbraid.spin.model.TemplateCall;
import org.topbraid.spin.model.Variable;
import org.topbraid.spin.model.print.PrintContext;
import org.topbraid.spin.model.print.Printable;
import org.topbraid.spin.model.print.StringPrintContext;
import org.topbraid.spin.system.SPINModuleRegistry;
import org.topbraid.spin.vocabulary.SPIN;

import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QuerySolutionMap;
import com.hp.hpl.jena.rdf.model.Literal;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;


/**
 * Some static util methods for SPIN that don't fit anywhere else.
 * 
 * @author Holger Knublauch
 */
public class SPINUtil {
	
	/**
	 * The name of the variable that will be used in type binding
	 * triple patterns (?this rdf:type ?TYPE_CLASS)
	 */
	public final static String TYPE_CLASS_VAR_NAME = "TYPE_CLASS";
	

	/**
	 * Collects all queries or template calls at a given class.
	 * @param cls  the class to get the queries at
	 * @param predicate  the predicate such as <code>spin:rule</code>
	 * @param results  the List to add the results to
	 */
	public static void addQueryOrTemplateCalls(Resource cls, Property predicate, List<QueryOrTemplateCall> results) {
		List<Statement> ss = JenaUtil.getStatementsList(JenaUtil.listAllProperties(cls, predicate));
		
		// Special case: we might have an instance of a template call like spl:Attribute
		//               Then try to find the Template in the registry
		if(ss.isEmpty() && cls != null && cls.isURIResource()) {
			Template template = SPINModuleRegistry.get().getTemplate(cls.getURI(), null);
			if(template != null) {
				ss = JenaUtil.getStatementsList(JenaUtil.listAllProperties(template, predicate));
			}
		}
		
		for(Statement s : ss) {
			if(s.getObject().isResource()) {
				TemplateCall templateCall = SPINFactory.asTemplateCall(s.getResource());
				if(templateCall != null) {
					results.add(new QueryOrTemplateCall(cls, templateCall));
				}
				else {
					Query query = SPINFactory.asQuery(s.getResource());
					if(query != null) {
						results.add(new QueryOrTemplateCall(cls, query));
					}
				}
			}
		}
	}
	
	
	/**
	 * Inserts a statement  ?this a ?TYPE_CLASS .  after the WHERE { keyword. 
	 * @param sb  the StringBuilder to manipulate
	 * @return the variable name of the ?TYPE_CLASS variable
	 */
	public static String addThisTypeClause(StringBuilder sb) {
		String varName = TYPE_CLASS_VAR_NAME;
		int index = sb.indexOf("WHERE {") + 7;
		sb.insert(index, " ?this a ?" + varName + " . ");
		return varName;
	}
	

	/**
	 * Applies variable bindings, replacing the values of one map with
	 * the values from a given variables map.
	 * @param map  the Map to modify
	 * @param bindings  the current variable bindings
	 */
	public static void applyBindings(Map<Property,RDFNode> map, Map<String,RDFNode> bindings) {
		for(Property property : new ArrayList<Property>(map.keySet())) {
			RDFNode value = map.get(property);
			Variable var = SPINFactory.asVariable(value);
			if(var != null) {
				String varName = var.getName();
				RDFNode b = bindings.get(varName);
				if(b != null) {
					map.put(property, b);
				}
			}
		}
	}
	

	/**
	 * Binds the variable ?this with a given value.
	 * @param qexec  the QueryExecution to modify
	 * @param value  the value to bind ?this with
	 */
	public static void bindThis(QueryExecution qexec, RDFNode value) {
		if(value != null) {
			QuerySolutionMap bindings = new QuerySolutionMap();
			bindings.add(SPIN.THIS_VAR_NAME, value);
			qexec.setInitialBinding(bindings);
		}
	}
	

	/**
	 * Checks whether a given query mentions the variable ?this anywhere.
	 * This can be used to check whether ?this needs to be bound before
	 * execution, etc.
	 * @param query  the query to test
	 * @return true  if query mentions ?this
	 */
	public static boolean containsThis(Query query) {
		return new ContainsVarChecker().contains(query, SPIN._this);
	}
	
	
	/**
	 * Attempts to convert a given RDFNode to a String so that it can be parsed into
	 * a Jena query object.  The node must be either a string Literal, or a sp:Query node
	 * or a template call.  If it's a template call then the resulting query string will
	 * have additional LET statements to "hard-bind" the template variables. 
	 * @param node  the RDFNode to convert
	 * @param usePrefixes  true to use qname abbreviations
	 * @return the String representation of node
	 * @throws IllegalArgumentException  if the node is not a valid SPIN Query or a String
	 */
	public static String getQueryString(RDFNode node, boolean usePrefixes) {
		if(node.isLiteral()) {
			return ((Literal)node).getLexicalForm();
		}
		else {
			Resource resource = (Resource)node;
			org.topbraid.spin.model.Query spinQuery = SPINFactory.asQuery(resource);
			if(spinQuery != null) {
				StringPrintContext p = new StringPrintContext();
				p.setUsePrefixes(usePrefixes);
				spinQuery.print(p);
				return p.getString();
			}
			else {
				TemplateCall templateCall = SPINFactory.asTemplateCall(resource);
				if(templateCall != null) {
					return templateCall.getQueryString();
				}
				else {
					throw new IllegalArgumentException("Node must be either literal or a SPIN query or a SPIN template call");
				}
			}
		}
	}
	
	
	/**
	 * Gets a Collection of all query strings defined as values of a given property.
	 * This will accept strings or SPIN expressions (including template calls).
	 * The query model is the subject's getModel().
	 * All sub-properties of property from the query model will also be queried.
	 * @param subject  the subject to get the values of
	 * @param property  the property to query
	 * @return a Set of query strings
	 */
	public static Set<String> getQueryStrings(Resource subject, Property property) {
		if(subject != null) {
			property = subject.getModel().getProperty(property.getURI());
		}
		Set<String> queryStrings = new HashSet<String>();
		Set<Resource> ps = JenaUtil.getAllSubProperties(property);
		ps.add(property);
		for(Resource p : ps) {
			StmtIterator it = property.getModel().listStatements(subject, JenaUtil.asProperty(p), (RDFNode)null);
			while(it.hasNext()) {
				RDFNode object = it.nextStatement().getObject();
				String str = getQueryString(object, false);
				queryStrings.add(str);
			}
		}
		return queryStrings;
	}
	
	
	public static Set<Resource> getURIResources(Printable query) {
		final Set<Resource> results = new HashSet<Resource>();
		StringPrintContext context = new StringPrintContext() {

			@Override
			public PrintContext clone() {
				return this;
			}

			@Override
			public void printURIResource(Resource resource) {
				super.printURIResource(resource);
				results.add(resource);
			}
		};
		query.print(context);
		return results;
	}
	
	
	/**
	 * Converts a map from Properties to RDFNode values to a Map from variable
	 * names (Strings) to those values, for quicker look up.
	 * @param map  the old Map
	 * @return the new Map
	 */
	public static Map<String,RDFNode> mapProperty2VarNames(Map<Property,RDFNode> map) {
		Map<String,RDFNode> results = new HashMap<String,RDFNode>();
		for(Property predicate : map.keySet()) {
			RDFNode value = map.get(predicate);
			results.put(predicate.getLocalName(), value);
		}
		return results;
	}
}
