package org.topbraid.spin.util;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.topbraid.spin.vocabulary.SPIN;
import org.topbraid.spin.vocabulary.SPL;

import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.vocabulary.RDF;


/**
 * Utilities related to the spl namespace.
 * 
 * @author Holger Knublauch
 */
public class SPLUtil {
	
	private static void addDefaultValuesForType(Resource cls, Map<Property,RDFNode> results, Set<Resource> reached) {
		
		reached.add(cls);
		
		StmtIterator it = cls.listProperties(SPIN.rule);
		while(it.hasNext()) {
			Statement s = it.nextStatement();
			if(s.getObject().isResource()) {
				Resource templateCall = s.getResource();
				if(templateCall.hasProperty(RDF.type, SPL.InferDefaultValue)) {
					Statement predicateS = templateCall.getProperty(SPL.predicate);
					if(predicateS != null && predicateS.getObject().isURIResource()) {
						Property predicate = cls.getModel().getProperty(predicateS.getResource().getURI());
						if(!results.containsKey(predicate)) {
							Statement v = templateCall.getProperty(SPL.defaultValue);
							if(v != null) {
								results.put(predicate, v.getObject());
							}
						}
					}
				}
			}
		}
		
		for(Resource superClass : JenaUtil.getSuperClasses(cls)) {
			if(!reached.contains(superClass)) {
				addDefaultValuesForType(superClass, results, reached);
			}
		}
	}
	
	
	private static RDFNode getDefaultValueForType(Resource cls, Property predicate, Set<Resource> reached) {
		reached.add(cls);
		StmtIterator it = cls.listProperties(SPIN.rule);
		while(it.hasNext()) {
			Statement s = it.nextStatement();
			if(s.getObject().isResource()) {
				Resource templateCall = s.getResource();
				if(templateCall.hasProperty(RDF.type, SPL.InferDefaultValue)) {
					if(templateCall.hasProperty(SPL.predicate, predicate)) {
						Statement v = templateCall.getProperty(SPL.defaultValue);
						if(v != null) {
							return v.getObject();
						}
					}
				}
			}
		}
		
		for(Resource superClass : JenaUtil.getSuperClasses(cls)) {
			if(!reached.contains(superClass)) {
				RDFNode value = getDefaultValueForType(superClass, predicate, reached);
				if(value != null) {
					return value;
				}
			}
		}
		
		return null;
	}
	
	
	/**
	 * Creates a Map from Properties to RDFNodes based on declared
	 * spl:InferDefaultValues.
	 * @param subject
	 * @return a Map from Properties to their default values (no null values)
	 */
	public static Map<Property,RDFNode> getDefaultValues(Resource subject) {
		Map<Property,RDFNode> results = new HashMap<Property,RDFNode>();
		Set<Resource> reached = new HashSet<Resource>();
		for(Resource type : JenaUtil.getTypes(subject)) {
			addDefaultValuesForType(type, results, reached);
		}
		return results;
	}


	/**
	 * Same as <code>getObject(subject, predicate, false)</code>.
	 * @see #getObject(Resource, Property, boolean)
	 */
	public static RDFNode getObject(Resource subject, Property predicate) {
		return getObject(subject, predicate, false);
	}
	

	/**
	 * Gets the (first) value of a subject/predicate combination.
	 * If no value exists, then it checks whether any spl:InferDefaultValue
	 * has been defined for the type(s) of the subject.
	 * No need to run inferences first.
	 * @param subject  the subject to get the object of
	 * @param predicate  the predicate
	 * @param includeSubProperties  true to also check for sub-properties of predicate
	 * @return the object or null
	 */
	public static RDFNode getObject(Resource subject, Property predicate, boolean includeSubProperties) {
		Statement s = subject.getProperty(predicate);
		if(s != null) {
			return s.getObject();
		}
		else {
			Set<Resource> reached = new HashSet<Resource>();
			for(Resource type : JenaUtil.getTypes(subject)) {
				RDFNode object = getDefaultValueForType(type, predicate, reached);
				if(object != null) {
					return object;
				}
			}
			if(includeSubProperties) {
				for(Resource subProperty : JenaUtil.getAllSubProperties(predicate)) {
					Property pred = (Property) subProperty.as(Property.class);
					RDFNode value = getObject(subject, pred, false);
					if(value != null) {
						return value;
					}
				}
			}
			
			return null;
		}
	}
}
